/*----------------------------------------------------------------------
  File    : citrie.c
  Contents: management of a trie (prefix tree)
            for conditional independence test results
  Author  : Christian Borgelt
  History : 2006.03.05 file created
            2006.03.06 conditioned pairs information stored directly
            2006.03.08 bug in function cit_add() fixed
            2013.09.04 adapted to definition of ATTID
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "citrie.h"
#include "arrays.h"
#ifdef STORAGE
#include "storage.h"
#endif

#ifdef _MSC_VER
#ifndef isnan
#define isnan(x)    _isnan(x)
#endif                          /* check for 'not a number' */
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

CITRIE* cit_create (ATTID cnt)
{                               /* --- create a cond. indep. trie */
  size_t size;                  /* size of the root table */
  CITRIE *cit;                  /* created cond. indep. trie */

  assert(cnt >= 0);             /* check the function argument */
  size = (size_t)cnt *(size_t)(cnt-1)/2;
  cit  = (CITRIE*)calloc(1, sizeof(CITRIE)
                       +(size-1)     *sizeof(CINODE*)
                       +(size_t) cnt *sizeof(ATTID));
  if (!cit) return NULL;        /* create the base structure */
  cit->cnt   = cnt;             /* and initialize it */
  cit->cicnt = 0;
  cit->cimax = 0;
  cit->vals  = NULL;
  cit->buf   = (ATTID*)(cit->trees +size);
  cit->msys  = ms_create(sizeof(CINODE), 2720);
  if (!cit->msys) { free(cit); return NULL; }
  return cit;                   /* return the created trie */
}  /* cit_create() */

/*--------------------------------------------------------------------*/

void cit_delete (CITRIE *cit)
{                               /* --- delete a cond. indep. trie */
  assert(cit);                  /* check the function arguments */
  if (cit->vals) free(cit->vals);
  ms_delete(cit->msys);         /* delete the node memory management */
  free(cit);                    /* and the base structure */
}  /* cit_delete() */

/*--------------------------------------------------------------------*/

static CINODE* create (CITRIE *cit, ATTID id, CINODE *succ)
{                               /* --- create a trie node */
  CINODE *node;                 /* created node */

  node = ms_alloc(cit->msys);   /* get a memory block for the node */
  if (!node) return NULL;       /* and initialize it */
  node->id   = id;   node->val   = NAN;
  node->succ = succ; node->child = NULL;
  return node;                  /* return the created node */
}  /* create() */

/*--------------------------------------------------------------------*/

int cit_add (CITRIE *cit, ATTID i, ATTID k, ATTID *cons, ATTID n,
             double val)
{                               /* --- add cond. indep. to trie */
  CINODE *node, *prec = NULL;   /* to traverse the trie nodes */

  assert(cit                    /* check the function arguments */
  &&    (i >= 0) && (i < cit->cnt)
  &&    (k >= 0) && (k < cit->cnt) && (i != k)
  &&    (n <= cit->cnt-2) && (cons || (n <= 0)) && (val >= 0.0));
  if (i > k) i = i*(i-1)/2 +k;  /* compute the table index */
  else       i = k*(k-1)/2 +i;  /* (linearize the 2-dim. index) */
  if (n <= 0) {                 /* if there are no conditions */
    if (!cit->vals) {           /* if there is no evaluation vector */
      k = cit->cnt *(cit->cnt-1) /2;
      cit->vals = (double*)malloc((size_t)k *sizeof(double));
      if (!cit->vals) return -1;
      while (--k >= 0) cit->vals[k] = NAN;
    }                           /* create an evaluation vector */
    cit->vals[i] = val;         /* store the evaluation */
    cit->cicnt++;               /* in the evaluation vector, */
    return 0;                   /* count it, and return 'ok' */
  }
  memcpy(cit->buf, cons, (size_t)n *sizeof(ATTID));
  aid_qsort(cit->buf, (size_t)n, +1);  /* copy and sort conditions */
  node = cit->trees[i];         /* get the root node for the pair */
  if (!node                     /* if the tree is empty or node */
  ||  (node->id > cit->buf[0])) {      /* must precede the root */
    cit->trees[i] = node = create(cit, cit->buf[0], node);
    if (!node) return -1;       /* create a new root node */
  }                             /* and go to the new node */
  for (i = 0; 1; ) {            /* traverse the attribute ids. */
    while (node && (node->id < cit->buf[i])) {
      prec = node; node = node->succ; }
    if (!node || (node->id > cit->buf[i])) {
      prec->succ = node = create(cit, cit->buf[i], node);
      if (!node) return -1;     /* create a new node and */
    }                           /* insert it into the sibling list */
    if (++i >= n) {             /* if all ids. have been matched */
      if (n > cit->cimax) cit->cimax = n;
      node->val = val; cit->cicnt++;
      return 0;                 /* store and count the evaluation */
    }                           /* and abort the function */
    prec = node; node = node->child;
    if (node) continue;         /* check for a sibling list */
    prec->child = node = create(cit, cit->buf[i], NULL);
    if (!node) return -1;       /* create a new node (first child) */
  }                             /* thus starting the sibling list */
}  /* cit_add() */

/*--------------------------------------------------------------------*/

double cit_get (CITRIE *cit, ATTID i, ATTID k, ATTID *cons, ATTID n)
{                               /* --- get cond. indep. from trie */
  CINODE *node;                 /* to traverse the trie nodes */

  assert(cit                    /* check the function arguments */
  &&    (i >= 0) && (i < cit->cnt)
  &&    (k >= 0) && (k < cit->cnt) && (i != k)
  &&    (n <= cit->cnt-2) && (cons || (n <= 0)));
  if (i > k) i = i*(i-1)/2 +k;  /* compute the table index */
  else       i = k*(k-1)/2 +i;  /* (linearize the 2-dim. index) */
  if (n <= 0)                   /* if there are no conditions */
    return (cit->vals) ? cit->vals[i] : NAN;
  memcpy(cit->buf, cons, (size_t)n *sizeof(ATTID));
  aid_qsort(cit->buf, (size_t)n, +1); /* copy and sort conditions */
  node = cit->trees[i];         /* get the root node for the pair */
  for (i = 0; 1; ) {            /* traverse the attribute ids. */
    if (!node || (node->id > cit->buf[i]))
      return NAN;               /* if no corresp. node found, abort */
    if (node->id < cit->buf[i]) /* if possibly before corresp. node, */
      node = node->succ;        /* go to the next node in the list */
    else {                      /* if at the corresponding node */
      if (++i >= n) return node->val;
      node = node->child;       /* if all identifiers were matched, */
    }                           /* return the stored evaluation, */
  }                             /* otherwise go to the child node */
}  /* cit_get() */

/*--------------------------------------------------------------------*/

static void desc (CINODE *node, ATTID *buf, ATTID d, FILE *file)
{                               /* --- describe evaluations */
  ATTID i;                      /* loop variable */

  for ( ; node; node = node->succ) {
    buf[d] = node->id;          /* traverse the sibling list */
    if (node->val >= 0.0) {     /* if there is an evaluation */
      fprintf(file, "%2"ATTID_FMT" %2"ATTID_FMT" |", buf[0], buf[1]);
      for (i = 2; i <= d; i++)  /* print conditioned variables and */
        fprintf(file, " %"ATTID_FMT, buf[i]);/* list of conditions */
      fprintf(file, "   : %g\n", node->val);
    }                           /* print evaluation */
    if (node->child) desc(node->child, buf, d+1, file);
  }                             /* process subtree recursively */
}  /* desc() */

/*--------------------------------------------------------------------*/

void cit_desc (CITRIE *cit, FILE *file)
{                               /* --- describe all evaluations */
  ATTID i, k, n;                /* loop variables */

  for (i = 0; i < cit->cnt; i++) {
    for (k = 0; k < i; k++) {   /* traverse all attribute pairs */
      n = i*(i-1)/2 +k;         /* get linearized index */
      if (cit->vals) {          /* print conditionless evaluation */
        fprintf(file, "%2"ATTID_FMT" %2"ATTID_FMT, i, k);
        fprintf(file, "   : %g\n", cit->vals[n]);
      }
      if (!cit->trees[n]) continue;
      cit->buf[0] = i; cit->buf[1] = k;
      desc(cit->trees[n], cit->buf, 2, file);
    }                           /* recursively print */
  }                             /* the conditional independences */
}  /* cit_desc() */

/*--------------------------------------------------------------------*/

static void stats (CINODE *node, ATTID *cnts)
{                               /* --- recursively collect statistics */
  for ( ; node; node = node->succ) {
    if (!isnan(node->val)) (*cnts)++;
    if (node->child) stats(node->child, cnts+1);
  }                             /* count the test of the node and */
}  /* stats() */                /* process children recursively */

/*--------------------------------------------------------------------*/

void cit_stats (CITRIE *cit, FILE *file)
{                               /* --- print statistics */
  size_t i, n;                  /* loop variables */

  assert(cit && file);          /* check the function arguments */
  memset(cit->buf, 0, (size_t)cit->cnt *sizeof(ATTID));
  n = (size_t)cit->cnt *(size_t)(cit->cnt-1) /2;
  if (cit->vals) {              /* if there are order 0 tests, */
    for (i = 0; i < n; i++)     /* traverse the variable pairs */
      if (!isnan(cit->vals[i])) cit->buf[0]++;
  }                             /* count the order 0 tests */
  for (i = 0; i < n; i++)       /* count the higher order tests */
    if (cit->trees[i]) stats(cit->trees[i], cit->buf +1);
  for (i = 0; i <= (size_t)cit->cimax; i++) {
    fprintf(file, "order %2"SIZE_FMT,  i);
    fprintf(file, ": %3"ATTID_FMT"\n", cit->buf[i]);
  }                             /* print counters per test order */
  fprintf(file, "total   : %3"SIZE_FMT"\n", cit->cicnt);
}  /* cit_stats() */            /* print total number of tests */

/*--------------------------------------------------------------------*/
#ifdef CIT_MAIN

int main (void)
{                               /* --- main function for testing */
  CITRIE *cit;                  /* created cond. indep. trie */
  ATTID  buf[8];                /* buffer for conditions */

  cit = cit_create(8);          /* create a cond. indep. trie */
  if (!cit) return -1;          /* for 8 attributes */
  buf[0] = 6; buf[1] = 4; buf[2] = 5;
  if (cit_add(cit, 1, 2, buf, 3, 1) != 0) return -1;
  printf("added 1 _||_2 | 6,4,5: 1\n");
  printf("1 _||_2 | *    : %g\n", cit_get(cit, 1, 2, buf, 0));
  printf("1 _||_2 | 6,4  : %g\n", cit_get(cit, 1, 2, buf, 2));
  printf("1 _||_2 | 6,4,5: %g\n", cit_get(cit, 1, 2, buf, 3));
  if (cit_add(cit, 1, 2, buf, 2, 2) != 0) return -1;
  printf("added 1 _||_2 | 6,4  : 2\n");
  printf("1 _||_2 | 6,4  : %g\n", cit_get(cit, 1, 2, buf, 2));
  buf[0] = 4; buf[1] = 5;
  if (cit_add(cit, 1, 2, buf, 2, 3) != 0) return -1;
  printf("added 1 _||_2 | 4,5  : 3\n");
  printf("1 _||_2 | 4,5  : %g\n", cit_get(cit, 1, 2, buf, 2));
  buf[0] = 3; buf[1] = 4;
  if (cit_add(cit, 1, 2, buf, 2, 4) != 0) return -1;
  printf("added 1 _||_2 | 3,4  : 4\n");
  printf("1 _||_2 | 3,4  : %g\n", cit_get(cit, 1, 2, buf, 2));
  buf[0] = 6; buf[1] = 4; buf[2] = 5;
  printf("1 _||_2 | 6,4  : %g\n", cit_get(cit, 1, 2, buf, 2));
  buf[0] = 4; buf[1] = 5;
  printf("1 _||_2 | 4,5  : %g\n", cit_get(cit, 1, 2, buf, 2));
  buf[0] = 3; buf[1] = 4;
  printf("1 _||_2 | 3,4  : %g\n", cit_get(cit, 1, 2, buf, 2));
  if (cit_add(cit, 1, 2, buf, 0, 5) != 0) return -1;
  printf("added 1 _||_2 | 3,4  : 5\n");
  printf("1 _||_2 | *    : %g\n", cit_get(cit, 1, 2, buf, 0));
  printf("statistics:\n");
  cit_stats(cit, stdout);
  return 0;                     /* return 'ok' */
}  /* main() */

#endif
