/*----------------------------------------------------------------------
  File    : gendb.c
  Contents: program to generate a random database
            which satisfies a given probability distribution
  Author  : Christian Borgelt
  History : 1995.10.19 file created
            1995.10.26 attribute set added
            1995.10.30 instantiation of attribute set added
            1995.11.26 adapted to modified attset functions
            1996.02.24 adapted to modified attset functions
            1996.03.10 alignment (option -a) added
            1996.03.17 adapted to modified parse function
            1996.04.09 adapted to modified gm_rand
            1996.11.22 options -b, -f, and -r added
            1998.06.23 adapted to modified attset functions
            1998.08.07 option -w (do not write header) added
            1999.04.18 simplified using the new module 'io'
            1999.04.24 distribution expansion (option -x) added
            2001.11.25 adapted to modified scanner functions
            2002.01.21 adapted to modified gramod and ptree modules
            2002.02.04 check for empty attribute domains added
            2003.08.16 slight changes in error message output
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include <time.h>
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_WRITE
#define AS_WRITE
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#ifndef GM_PARSE
#define GM_PARSE
#endif
#include "gramod.h"
#include "random.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "gendb"
#define DESCRIPTION "generate a random database"
#define VERSION     "version 3.1 (2014.10.24)         " \
                    "(c) 1995-2014   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse error */
#define E_TPLCNT    (-10)       /* invalid number of tuples */
#define E_EMPTY     (-11)       /* empty attribute domain */
#define E_LOOP      (-12)       /* loop in conditions */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_TPLCNT  -10 */  "invalid number of tuples",
  /* E_EMPTY   -11 */  "domain of attribute %s is empty",
  /* E_LOOP    -12 */  "condition loop detected",
  /*           -13 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR    *prgname;       /* program name for error messages */
static SCANNER  *scan   = NULL; /* scanner (for graphical model) */
static ATTSET   *attset = NULL; /* attribute set */
static GRAMOD   *gramod = NULL; /* graphical model (Bayesian network) */
static TABWRITE *twrite = NULL; /* table writer */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (gramod) gm_delete(gramod,  0); \
  if (attset) as_delete(attset);     \
  if (scan)   scn_delete(scan,   1); \
  if (twrite) twr_delete(twrite, 1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counter */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_pd   = NULL;      /* name of prob. distrib. file */
  CCHAR   *fn_tab  = NULL;      /* name of table output file */
  CCHAR   *blank   = NULL;      /* blank  character */
  CCHAR   *fldsep  = NULL;      /* field  separator */
  CCHAR   *recsep  = NULL;      /* record separator */
  int     mout     = AS_ATT;    /* table file write mode */
  int     expand   = 0;         /* whether to expand the distribution */
  size_t  tplcnt   = 1000, n;   /* number of tuples to generate */
  long    seed;                 /* seed for random number generator */
  double  prob;                 /* probability of a tuple */
  ATT     *att;                 /* to traverse the attributes */
  INST    *inst;                /* to traverse the att. instances */
  ATTID   m, c;                 /* number of attributes */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */
  seed    = (long)time(NULL);   /* and get a default seed value */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] pdfile tabfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-n#      number of tuples to generate           "
                    "(default: %"SIZE_FMT")\n", tplcnt);
    printf("-s#      seed for random number generator       "
                    "(default: time)\n");
    printf("-x       expand the distribution (do not sample)\n");
    printf("-a       align fields in output table           "
                    "(default: single separator)\n");
    printf("-w       do not write field names to the output file\n");
    printf("-r#      record separator                       "
                    "(default: \"\\n\")\n");
    printf("-f#      field  separator                       "
                    "(default: \" \")\n");
    printf("-b#      blank  character                       "
                    "(default: \" \")\n");
    printf("pdfile   file containing probability distributions\n");
    printf("tabfile  file to write generated tuples to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'n': tplcnt = (size_t)strtol(s, &s, 0); break;
          case 's': seed   =         strtol(s, &s, 0); break;
          case 'x': expand = 1;                        break;
          case 'w': mout  &= AS_ATT;                   break;
          case 'a': mout  |= AS_ALIGN;                 break;
          case 'b': optarg = &blank;                   break;
          case 'f': optarg = &fldsep;                  break;
          case 'r': optarg = &recsep;                  break;
          default : error(E_OPTION, *(--s));           break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-options */
        case  0: fn_pd  = s;      break;
        case  1: fn_tab = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note input/output filename */
    }
  }
  if (optarg) error(E_OPTARG);  /* check the option argument */
  if (k != 2) error(E_ARGCNT);  /* and the number of arguments */
  if (tplcnt <= 0) error(E_TPLCNT);
  if (expand) mout |= AS_WEIGHT;/* set the weight output flag */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- read probability distributions --- */
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  scan = scn_create();          /* create a scanner */
  if (!scan)   error(E_NOMEM);  /* for the classifier file */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_pd) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  if (as_parse(attset, scan, AT_ALL, 1) != 0)
    error(E_PARSE, scn_name(scan)); /* parse domain descriptions */
  gramod = gm_parse(attset, scan, GM_PROB|GM_ALL);
  if (!gramod || !scn_eof(scan, 1)) /* parse the distributions */
    error(E_PARSE, scn_name(scan)); /* as a graphical model */
  if (gm_check(gramod, -1, 0) != 0)
    error(E_LOOP);              /* check conditions to detect loops */
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- generate database --- */
  twrite = twr_create();        /* create a table writer and */
  if (!twrite) error(E_NOMEM);  /* configure the characters */
  twr_xchars(twrite, recsep, fldsep, blank, "");
  t = clock();                  /* start timer, open output file */
  if (twr_open(twrite, NULL, fn_tab) != 0)
    error(E_FOPEN, twr_name(twrite));
  fprintf(stderr, "writing %s ... ", twr_name(twrite));
  if ((mout & AS_ATT)           /* write the attributes names */
  &&  (as_write(attset, twrite, mout) != 0))
    error(E_FWRITE, twr_name(twrite));
  mout = AS_INST | (mout & ~AS_ATT);
  rseed((unsigned)seed);        /* init. random number generator */
  if (expand) {                 /* if to expand the distribution */
    for (c = m = as_attcnt(attset); --c >= 0; ) {
      att = as_att(attset, c);  /* traverse the attributes */
      att_inst(att)->n = att_valcnt(att) -1;
    }                           /* set the last value of each domain */
    do {                        /* generate tuples */
      prob = gm_exec(gramod);   /* determine tuple probability */
      as_setwgt(attset, (WEIGHT)(prob *(double)tplcnt));
      if (as_write(attset, twrite, mout) != 0)
        error(E_FWRITE, twr_name(twrite)); /* write generated tuple */
      for (c = m; --c >= 0; ) {    /* traverse the attributes */
        att  = as_att(attset, c);  /* and their instances and */
        inst = att_inst(att);      /* set the next attribute value */
        if (--inst->n >= 0) break; /* or  the last attribute value */
        inst->n = att_valcnt(att) -1;
      }                         /* (compute next value combination) */
    } while (c >= 0); }         /* while there is another combination */
  else {                        /* if to sample from the distrib. */
    rseed((unsigned)seed);      /* init. random number generator */
    for (n = 0; n < tplcnt; n++) {     /* generate random tuples */
      gm_rand(gramod, drand);   /* instantiate the attribute set */
      if (as_write(attset, twrite, mout) != 0)
        error(E_FWRITE, twr_name(twrite));
    }                           /* write the generated tuple */
  }                             /* to the output file */
  if (twr_close(twrite) != 0) error(E_FWRITE, twr_name(twrite));
  twr_delete(twrite, 1);        /* close the output file and */
  twrite = NULL;                /* delete the table writer */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"SIZE_FMT" tuple(s)]",      n);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
