/*----------------------------------------------------------------------
  File    : ptree.h
  Contents: probability/possibility tree management
  Author  : Christian Borgelt
  History : 1995.10.22 file created
            1995.10.26 PTREE.curr and pt_down() and pt_up() added
            1995.10.27 range checks removed
            1995.10.29 function pt_conadd() added
            1995.11.02 functions pt_conrem() and pt_eval() added
            1995.11.08 function pt_getcnt() added
            1995.12.11 evaluation measure 'EM_INFGAIN' added
            1996.02.05 some evaluation measures added
            1996.02.07 evaluation measure 'EM_GINI' added
            1996.02.23 functions pt_mname() and pt_minfo() added
            1996.03.15 number of levels restricted to PT_MAXCON+1
            1996.06.29 second specificity gain measure added
            1996.11.15 function pt_total() added
            1996.11.27 type of counters changed to floating point
            1997.02.28 tree mode (PT_PROB/PT_POSS) added, redesign
            1997.03.01 functions pt_aggr() and pt_conrem() improved
            1997.03.06 some evaluation measures added
            1997.03.10 second sym. info./spec. gain ratio added
            1997.03.11 alternate measure versions added
            1997.03.25 function pt_depend() added
            1998.02.13 links between nodes made possible (pt_link)
            1998.02.18 some evaluation measures added
            1998.02.20 Bayesian Dirichlet l.e. uniform metric added
            1998.06.10 semantics of function pt_down() changed
            1999.02.20 structure PTLVL added, structure PTREE modified
            1999.02.24 based on attribute set management
            1999.04.25 function pt_concopy() added
            1999.12.11 evaluation measure 'EM_WDIFF' added
            2001.11.01 module renamed to ptree (prob./poss. tree)
            2002.02.24 function pt_getcntx() added
            2002.03.05 description flag PT_INFO added
            2002.03.11 number of tuples added for poss. normalization
            2002.04.04 function pt_parcnt() added
            2002.04.10 parameter flags for function pt_parse removed
            2002.07.01 field pathcnt added to structure PTLEAF
            2002.07.03 field occur added to structure PTREE
            2002.07.16 parameter 'delnodes' added to function pt_clear
            2002.07.17 PTREE.buf added (for nonspecificity comp.)
            2006.02.16 restriction of number of conditions removed
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#ifndef __PTREE__
#define __PTREE__
#include <limits.h>
#include <float.h>
#include "fntypes.h"
#ifdef PT_PARSE
#ifndef SCN_SCAN
#define SCN_SCAN
#endif
#include "scanner.h"
#endif
#include "attset.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- tree types --- */
#define PT_PROB     0x0000      /* probability tree (sum) */
#define PT_POSS     0x0001      /* possibility tree (maximum) */
#define PT_CITEST   0x0002      /* cond. independence test */
#define PT_AUTO     0x8000      /* determine type from input file */

/* --- node flags --- */
#define PT_OCCUR    INT_MIN     /* whether a subtree is not empty */

/* --- modes for pt_up --- */
#define PT_ROOT      0          /* go up to the root node */
#define PT_PARENT    1          /* go up to the true parent */
#define PT_DESCENT   2          /* go up to where descended from */

/* --- modes for pt_desc/pt_parse --- */
#define PT_TITLE    0x0001      /* print a title (as a comment) */
#define PT_INFO     0x0002      /* print add. info (as a comment) */
#define PT_FULL     0x0004      /* print a full tree (all branches) */
#define PT_REL      0x0008      /* print relative frequencies */
#define PT_INDENT   0x0010      /* indent lines (for graph. model) */

/* --- probabilistic evaluation measures --- */
#define PT_NONE      0          /* no measure */
#define PT_INFGAIN   1          /* information gain */
#define PT_INFGBAL   2          /* balanced information gain */
#define PT_INFGR     3          /* information gain ratio */
#define PT_INFSGR1   4          /* sym. information gain ratio 1 */
#define PT_INFSGR2   5          /* sym. information gain ratio 2 */
#define PT_QIGAIN    6          /* quadratic information gain */
#define PT_QIGBAL    7          /* balanced quad. information gain */
#define PT_QIGR      8          /* quad. information gain ratio */
#define PT_QISGR1    9          /* sym. quad. info. gain ratio 1 */
#define PT_QISGR2   10          /* sym. quad. info. gain ratio 2 */
#define PT_GINI     11          /* gini index */
#define PT_GINISYM  12          /* symmetric gini index */
#define PT_GINIMOD  13          /* modified gini index */
#define PT_RELIEF   14          /* relief measure */
#define PT_WDIFF    15          /* sum of weighted differences */
#define PT_CHI2     16          /* chi^2 measure */
#define PT_CHI2NRM  17          /* normalized chi^2 measure */
#define PT_WEVID    18          /* weight of evidence */
#define PT_RELEV    19          /* relevance */
#define PT_BDM      20          /* Bayesian-Dirichlet / K2 metric */
#define PT_BDMOD    21          /* modified BD / K2 metric */
#define PT_RDLREL   22          /* red. of desc. length (rel. freq.) */
#define PT_RDLABS   23          /* red. of desc. length (abs. freq.) */
#define PT_STOCO    24          /* stochastic complexity */
#define PT_UNKNOWN  25          /* unknown measure */

/* --- possibilistic evaluation measures --- */
#define PT_SPCGAIN   1          /* specificity gain */
#define PT_SPCGBAL   2          /* balanced specificity gain */
#define PT_SPCGR     3          /* specificity gain ratio */
#define PT_SPCSGR1   4          /* sym. specificity gain ratio */
#define PT_SPCSGR2   5          /* sym. specificity gain ratio */
/*      PT_WDIFF    15 */       /* weighted general difference */
/*      PT_CHI2     16 */       /* possibilistic chi^2 measure */
#define PT_MUTSPC   20          /* mutual specificity */

/* --- other values --- */
#define PT_SYM      0x8000      /* whether measure is symmetric */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct ptnode {         /* --- prob./poss. tree node --- */
  struct ptnode *succ;          /* successor node on same level */
  struct ptnode *parent;        /* parent node on preceding level */
  VALID         index;          /* vector index in parent node */
  VALID         pathcnt;        /* number of paths to this node */
  struct ptnode *children[1];   /* vector of child nodes */
} PTNODE;                       /* (prob./poss. tree node) */

typedef struct ptleaf {         /* --- prob./poss. tree leaf --- */
  struct ptleaf *succ;          /* successor node on same level */
  struct ptnode *parent;        /* parent node on preceding level */
  VALID         index;          /* vector index in parent node */
  VALID         pathcnt;        /* number of paths to this leaf */
  double        eval;           /* evaluation term */
  double        total;          /* aggregate of leaf counters */
  WEIGHT        cnts[1];        /* vector of counters */
} PTLEAF;                       /* (prob./poss. tree leaf) */

typedef struct {                /* --- prob./poss. tree level --- */
  VALID  cnt;                   /* number of children per node */
  ATT    *att;                  /* corresponding attribute */
  PTNODE *list;                 /* list of nodes on this level */
  PTNODE *node;                 /* current node on this level */
  VALID  index;                 /* child index used in current node */
  VALID  link;                  /* child index used on link */
  double total;                 /* aggregate of counters in buffer */
  WEIGHT *buf;                  /* buffer for aggregation */
} PTLVL;                        /* (prob./poss. tree level) */

typedef struct {                /* --- prob./poss. tree --- */
  ATT    *att;                  /* conditioned attribute */
  int    type;                  /* tree type (PT_PROB or PT_POSS) */
  ATTID  concnt;                /* number of conditions */
  ATTID  convsz;                /* size of condition/level vector */
  PTLVL  *levels;               /* tree level descriptions */
  VALID  fullcnt;               /* number of leaves of a full tree */
  VALID  leafcnt;               /* number of leaf nodes */
  VALID  parcnt;                /* number of parameters */
  ATTID  pathlen;               /* path length (current level) */
  PTNODE *curr;                 /* current node (tree cursor) */
  double tplcnt;                /* number of tuples for norm. */
  double total;                 /* total of all counters */
  WEIGHT lcorr;                 /* Laplace correction */
  int    occur;                 /* whether occurrence flags are set */
  int    measure;               /* evaluation measure */
  double params[2];             /* evaluation parameters */
  double res0[5];               /* intermediate evaluation results */
  double res1[5];               /* ditto, for final computation */
  double eval;                  /* result of evaluation */
  WEIGHT *buf;                  /* buffer for specificity computation */
} PTREE;                        /* (prob./poss. tree) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern PTREE* pt_create  (ATT *att,  int type);
extern void   pt_delete  (PTREE *pt, int delatts);
extern void   pt_clear   (PTREE *pt, int delnodes);
extern int    pt_type    (PTREE *pt);
extern double pt_setnorm (PTREE *pt, double tplcnt);
extern double pt_getnorm (PTREE *pt);
extern WEIGHT pt_setlc   (PTREE *pt, WEIGHT lcorr);
extern WEIGHT pt_getlc   (PTREE *pt);

extern ATTID  pt_attid   (PTREE *pt);
extern ATT*   pt_att     (PTREE *pt);
extern ATTID  pt_concnt  (PTREE *pt);
extern int    pt_conadd  (PTREE *pt, ATTID lvlid, ATT *att);
extern void   pt_conrem  (PTREE *pt, ATTID lvlid, int delatt);
extern int    pt_concopy (PTREE *dst, PTREE *src);
extern ATTID  pt_conid   (PTREE *pt, ATTID lvlid);
extern ATT*   pt_con     (PTREE *pt, ATTID lvlid);

extern ATTID  pt_level   (PTREE *pt);
extern int    pt_atleaf  (PTREE *pt);
extern void   pt_down    (PTREE *pt, VALID index);
extern void   pt_up      (PTREE *pt, int mode);
extern int    pt_setcnt  (PTREE *pt, VALID index, WEIGHT cnt);
extern WEIGHT pt_getcnt  (PTREE *pt, VALID index);
extern WEIGHT pt_getcntx (PTREE *pt);
extern double pt_gettot  (PTREE *pt);
extern int    pt_occur   (PTREE *pt);
extern ATTID  pt_path    (PTREE *pt, VALID *path);
extern int    pt_link    (PTREE *pt, VALID index, const VALID *path);

extern VALID  pt_parcnt  (PTREE *pt);
extern double pt_total   (PTREE *pt);
extern int    pt_aggr    (PTREE *pt, WEIGHT wgt, int distuv);
extern VALID  pt_rand    (PTREE *pt, RANDFN randfn);
extern double pt_exec    (PTREE *pt);
extern double pt_poss    (PTREE *pt);

extern double pt_eval    (PTREE *pt, int measure, double *params);
extern double pt_reeval  (PTREE *pt);
extern double pt_local   (PTREE *pt, int lwise, double minimp);
extern double pt_depend  (PTREE *pt, int measure, double *params);
extern CCHAR* pt_mname   (int type, int measure);
extern int    pt_minfo   (int type, int measure);
extern double pt_nsp     (WEIGHT *dist, VALID n);

extern int    pt_desc    (PTREE *pt, FILE *file, int mode, int maxlen);
#ifdef PT_PARSE
extern PTREE* pt_parse   (ATTSET *attset, SCANNER *scan, int type);
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define pt_type(p)       ((p)->type)
#define pt_setnorm(p,c)  ((p)->tplcnt = (c))
#define pt_getnorm(p)    ((p)->tplcnt)
#define pt_setlc(p,c)    ((p)->lcorr = (c))
#define pt_getlc(p)      ((p)->lcorr)

#define pt_attid(p)      att_id((p)->att)
#define pt_att(p)        ((p)->att)
#define pt_concnt(p)     ((p)->concnt)
#define pt_conid(p,l)    att_id((p)->levels[l].att)
#define pt_con(p,l)      ((p)->levels[l].att)

#define pt_level(p)      ((p)->pathlen)
#define pt_atleaf(p)     ((p)->pathlen >= (p)->concnt)
#define pt_getcnt(p,i)   ((p)->curr ? ((PTLEAF*)p->curr)->cnts[i] : 0)
#endif
